@@ -120,9 +120,22 @@
     Updated by      Richard Gooch   25-DEC-1999: Discard lookups on "/dev/log"
   and "/dev/initctl". Added "MODLOAD" action.
 
-    Last updated by Richard Gooch   3-FEB-2000: Bitch and moan if modprobe(8)
+    Updated by      Richard Gooch   3-FEB-2000: Bitch and moan if modprobe(8)
   is used with "EXECUTE" action.
 
+    Updated by      Richard Gooch   3-MAR-2000: Bug fix with inclusion
+  directives.
+
+    Updated by      Richard Gooch   6-MAR-2000: Support negative UID and GID in
+  config file. Thanks to Chris Richards <crichard@wso.williams.edu>.
+
+    Updated by      Richard Gooch   12-APR-2000: More efficient discarding of
+  lookups on "/dev/log" and "/dev/initctl".
+
+    Last updated by Richard Gooch   16-APR-2000: Added "MFUNCTION" and
+  "CFUNCTION" actions. Added compatibility entries for Computone Multiport
+  serial devices.
+
 
 */
 #include <unistd.h>
@@ -157,7 +170,7 @@
 #endif
 
 
-#define VERSION "1.3.1"
+#define VERSION "1.3.3"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
@@ -173,11 +186,13 @@
 #define AC_PERMISSIONS              0
 #define AC_MODLOAD                  1
 #define AC_EXECUTE                  2
-#define AC_IGNORE                   3
-#define AC_MKOLDCOMPAT              4
-#define AC_MKNEWCOMPAT              5
-#define AC_RMOLDCOMPAT              6
-#define AC_RMNEWCOMPAT              7
+#define AC_MFUNCTION                3
+#define AC_CFUNCTION                4
+#define AC_IGNORE                   5
+#define AC_MKOLDCOMPAT              6
+#define AC_MKNEWCOMPAT              7
+#define AC_RMOLDCOMPAT              8
+#define AC_RMNEWCOMPAT              9
 
 #define MKACTION(what,when) (struct action_type) {what, when}
 
@@ -195,7 +210,19 @@ struct permissions_type
 
 struct execute_type
 {
-    char *argv[MAX_ARGS + 1];
+    char *argv[MAX_ARGS + 1];  /*  argv[0] must always be the programme  */
+};
+
+struct call_function_type
+{
+    struct shared_object *so;
+    union
+    {
+	int (*m) (int argc, char **argv);
+	int (*c) (void *arg1, void *arg2, void *arg3, void *arg4, void *arg5);
+    }
+    func;
+    char *argv[MAX_ARGS + 1];  /*  argv[0] is the function name          */
 };
 
 struct action_type
@@ -204,7 +231,7 @@ struct action_type
     unsigned int when;
 };
 
-struct config_entry
+struct config_entry_struct
 {
     struct action_type action;
     regex_t preg;
@@ -212,9 +239,10 @@ struct config_entry
     {
 	struct permissions_type permissions;
 	struct execute_type execute;
+	struct call_function_type function;
     }
     u;
-    struct config_entry *next;
+    struct config_entry_struct *next;
 };
 
 struct get_variable_info
@@ -224,6 +252,13 @@ struct get_variable_info
     char devpath[STRING_LENGTH];
 };
 
+struct shared_object
+{
+    CONST char *name;
+    void *handle;
+    struct shared_object *next;
+};
+
 
 /*  External functions  */
 EXTERN_FUNCTION (flag st_expr_expand,
@@ -240,11 +275,13 @@ static void read_config (CONST char *filename, flag optional,
 static void do_servicing (int fd, unsigned long event_mask);
 static void service_name (const struct devfsd_notify_struct *info);
 static void action_permissions (const struct devfsd_notify_struct *info,
-				const struct config_entry *entry);
+				const struct config_entry_struct *entry);
 static void action_modload (const struct devfsd_notify_struct *info,
-			    const struct config_entry *entry);
+			    const struct config_entry_struct *entry);
 static void action_execute (const struct devfsd_notify_struct *info,
-			    const struct config_entry *entry);
+			    const struct config_entry_struct *entry);
+static void action_call_function (const struct devfsd_notify_struct *info,
+				  const struct config_entry_struct *entry);
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action);
 static void free_config ();
@@ -257,11 +294,13 @@ static CONST char *get_variable (CONST char *variable, void *info);
 static void do_open_syslog ();
 static void do_scan_and_service (DIR *dp, CONST char *dirname);
 static flag make_symlink (const char *oldpath, const char *newpath);
+static struct shared_object *get_shared_object (CONST char *name);
 
 
 /*  Private data  */
-static struct config_entry *first_config = NULL;
-static struct config_entry *last_config = NULL;
+static struct config_entry_struct *first_config = NULL;
+static struct config_entry_struct *last_config = NULL;
+static struct shared_object *first_so = NULL;
 static const char *mount_point = NULL;
 static flag no_syslog = FALSE;
 static flag syslog_is_open = FALSE;
@@ -279,7 +318,7 @@ static struct initial_symlink_struct
     {"fd/2", "stderr"},
     {NULL, NULL},
 };
-    
+
 
 /*  Public functions follow  */
 
@@ -479,7 +518,7 @@ static void read_config (CONST char *filename, flag optional,
     while (fgets (buf, STRING_LENGTH, fp) != NULL)
     {
 	int err, num_args, count;
-	struct config_entry *new;
+	struct config_entry_struct *new;
 	char p[MAX_ARGS][STRING_LENGTH];
 	char when[STRING_LENGTH], what[STRING_LENGTH];
 	char name[STRING_LENGTH], tmp[STRING_LENGTH];
@@ -508,9 +547,9 @@ static void read_config (CONST char *filename, flag optional,
 	if ( (strcasecmp (when, "INCLUDE") == 0) ||
 	     (strcasecmp (when, "OPTIONAL_INCLUDE") == 0) )
 	{
-	    st_expr_expand (name, STRING_LENGTH, what, get_variable, NULL,
+	    st_expr_expand (name, STRING_LENGTH, name, get_variable, NULL,
 			    NULL);
-	    read_config (what, (toupper (when[0]) == 'I') ? FALSE : TRUE,
+	    read_config (name, (toupper (when[0]) == 'I') ? FALSE : TRUE,
 			 event_mask);
 	    continue;
 	}
@@ -585,6 +624,45 @@ static void read_config (CONST char *filename, flag optional,
 	    }
 	    new->u.execute.argv[num_args] = NULL;
 	}
+	else if ( (strcasecmp (what, "MFUNCTION") == 0) ||
+		  (strcasecmp (what, "CFUNCTION") == 0) )
+	{
+	    struct shared_object *so;
+
+	    new->action.what = (strcasecmp (what, "MFUNCTION") == 0) ?
+		AC_MFUNCTION : AC_CFUNCTION;
+	    num_args -= 3;
+	    if (num_args < 2)
+	    {
+		SYSLOG (LOG_ERR, "missing path and function in \"%s\"\n",line);
+		exit (1);
+	    }
+	    if ( ( so = get_shared_object (p[0]) ) == NULL )
+	    {
+		SYSLOG ( LOG_ERR, "error loading: \"%s\"\t%s\n",
+			 p[0], dlerror () );
+		exit (1);
+	    }
+	    if ( ( new->u.function.func.m = dlsym (so->handle,p[1]) ) == NULL )
+	    {
+		SYSLOG (LOG_ERR, "symbol: \"%s\" not found in %s\n",
+			p[1], p[0]);
+		exit (1);
+	    }
+	    --num_args;
+	    for (count = 0; count < num_args; ++count)
+	    {
+		err = strlen (p[count + 1]) + 1;
+		if ( ( new->u.function.argv[count] = malloc (err) ) == NULL )
+		{
+		    SYSLOG (LOG_ERR, "error allocating\n");
+		    SYSLOG (LOG_ERR, "exiting\n");
+		    exit (1);
+		}
+		strcpy (new->u.function.argv[count], p[count + 1]);
+	    }
+	    new->u.function.argv[num_args] = NULL;
+	}
 	else if (strcasecmp (what, "IGNORE") == 0)
 	    new->action.what = AC_IGNORE;
 	else if (strcasecmp (what, "MKOLDCOMPAT") == 0)
@@ -676,12 +754,18 @@ static void service_name (const struct devfsd_notify_struct *info)
 */
 {
     regmatch_t mbuf;
-    struct config_entry *entry;
+    struct config_entry_struct *entry;
 
     if (info->overrun_count > 0)
 	SYSLOG (LOG_ERR, "%u events have been lost!\n", info->overrun_count);
     if (trace_level > 2) fprintf (stderr, "Looking for \"%s\" (%d)\n",
 				  info->devname, info->type);
+    /*  Discard lookups on "/dev/log" and "/dev/initctl"  */
+    if (info->type == DEVFSD_NOTIFY_LOOKUP)
+    {
+	if ( (strcmp (info->devname, "log") == 0) ||
+	     (strcmp (info->devname, "initctl") == 0) ) return;
+    }
     for (entry = first_config; entry != NULL; entry = entry->next)
     {
 	if (trace_level > 3) fprintf (stderr, "\t\tProcess \"%s\" (%d)\n",
@@ -690,12 +774,6 @@ static void service_name (const struct devfsd_notify_struct *info)
 	 */
 	if (info->type != entry->action.when) continue;
 	if (regexec (&entry->preg, info->devname, 1, &mbuf, 0) != 0) continue;
-	/*  Discard lookups on "/dev/log" and "/dev/initctl"  */
-	if (info->type == DEVFSD_NOTIFY_LOOKUP)
-	{
-	    if ( (strcmp (info->devname, "log") == 0) ||
-		 (strcmp (info->devname, "initctl") == 0) ) continue;
-	}
 	switch (entry->action.what)
 	{
 	  case AC_PERMISSIONS:
@@ -707,6 +785,10 @@ static void service_name (const struct devfsd_notify_struct *info)
 	  case AC_EXECUTE:
 	    action_execute (info, entry);
 	    break;
+	  case AC_MFUNCTION:
+	  case AC_CFUNCTION:
+	    action_call_function (info, entry);
+	    break;
 	  case AC_IGNORE:
 	    return;
 	    /*break;*/
@@ -725,7 +807,7 @@ static void service_name (const struct devfsd_notify_struct *info)
 }   /*  End Function service_name  */
 
 static void action_permissions (const struct devfsd_notify_struct *info,
-				const struct config_entry *entry)
+				const struct config_entry_struct *entry)
 /*  [SUMMARY] Update permissions for a device entry.
     <info> The devfs change.
     <entry> The config file entry.
@@ -780,7 +862,7 @@ static void action_permissions (const struct devfsd_notify_struct *info,
 }   /*  End Function action_permissions  */
 
 static void action_modload (const struct devfsd_notify_struct *info,
-			    const struct config_entry *entry)
+			    const struct config_entry_struct *entry)
 /*  [SUMMARY] Load a module.
     <info> The devfs change.
     <entry> The config file entry.
@@ -832,7 +914,7 @@ static void action_modload (const struct devfsd_notify_struct *info,
 }   /*  End Function action_modload  */
 
 static void action_execute (const struct devfsd_notify_struct *info,
-			    const struct config_entry *entry)
+			    const struct config_entry_struct *entry)
 /*  [SUMMARY] Execute a programme.
     <info> The devfs change.
     <entry> The config file entry.
@@ -883,6 +965,58 @@ static void action_execute (const struct devfsd_notify_struct *info,
     SYSLOG (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
 }   /*  End Function action_execute  */
 
+static void action_call_function (const struct devfsd_notify_struct *info,
+				  const struct config_entry_struct *entry)
+/*  [SUMMARY] Call a function.
+    <info> The devfs change.
+    <entry> The config file entry.
+    [RETURNS] Nothing.
+*/
+{
+    unsigned int count;
+    struct get_variable_info gv_info;
+    char *argv[MAX_ARGS + 1];
+    char largv[MAX_ARGS + 1][STRING_LENGTH];
+
+    gv_info.info = info;
+    gv_info.devname = info->devname;
+    strcpy (gv_info.devpath, mount_point);
+    strcat (gv_info.devpath, "/");
+    strcat (gv_info.devpath, info->devname);
+    if (trace_level > 1)
+	fprintf (stderr, "Calling %s in %s for \"%s\" | \"%s\":",
+		 entry->u.function.argv[0], entry->u.function.so->name,
+		 info->devname, gv_info.devpath);
+    argv[0] = (char *) entry->u.function.so->name;
+    for (count = 1; entry->u.function.argv[count] != NULL; ++count)
+    {
+	if (strcmp (entry->u.function.argv[count], "INFO") == 0)
+	    argv[count] = (char *) info;
+	else if (strcmp (entry->u.function.argv[count], "ENTRY") == 0)
+	    argv[count] = (char *) entry;
+	else
+	{
+	    st_expr_expand (largv[count], STRING_LENGTH,
+			    entry->u.function.argv[count], get_variable,
+			    &gv_info, NULL);
+	    argv[count] = largv[count];
+	    if (trace_level > 1) fprintf (stderr, " '%s'", argv[count]);
+	}
+    }
+    if (trace_level > 1) fprintf (stderr, ". i.e. %d arg%s\n",
+				  count - 1, (count == 2) ? "" : "s");
+    argv[count] = NULL;
+    if (entry->action.what == AC_MFUNCTION)
+    {
+	if ( (*entry->u.function.func.m) (count, argv) == 0 )
+	    return;
+    }
+    if ( (*entry->u.function.func.c) (argv[1], argv[2], argv[3], argv[4],
+				      argv[5]) == 0 ) return;
+    SYSLOG (LOG_ERR, "error calling: \"%s\" in \"%s\"\n",
+	    entry->u.function.argv[0], argv[0]);
+}   /*  End Function action_call_function  */
+
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action)
 /*  [SUMMARY] Process a compatibility request.
@@ -1144,6 +1278,21 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (compat_buf, "ttyS%s", info->devname + 4);
 	    compat_name = compat_buf;
 	}
+	else if (strncmp (info->devname, "ip2/", 4) == 0)
+	{
+	    sprintf (compat_buf, "ip2%s", info->devname + 4);
+	    compat_name = compat_buf;
+	}
+	else if (strncmp (info->devname, "ttf/", 4) == 0)
+	{
+	    sprintf (compat_buf, "ttyF%s", info->devname + 4);
+	    compat_name = compat_buf;
+	}
+	else if (strncmp (info->devname, "cuf/", 4) == 0)
+	{
+	    sprintf (compat_buf, "cuf%s", info->devname + 4);
+	    compat_name = compat_buf;
+	}
 	else if (strncmp (info->devname, "vc/", 3) == 0)
 	{
 	    sprintf (compat_buf, "tty%s", info->devname + 3);
@@ -1345,26 +1494,45 @@ static void free_config ()
     [RETURNS] Nothing.
 */
 {
-    struct config_entry *entry, *next;
+    struct config_entry_struct *c_entry;
+    struct shared_object *so_entry;
+    void *next;
 
-    for (entry = first_config; entry != NULL; entry = next)
+    for (c_entry = first_config; c_entry != NULL; c_entry = next)
     {
 	unsigned int count;
 
-	next = entry->next;
-	regfree (&entry->preg);
-	if (entry->action.what == AC_EXECUTE)
+	next = c_entry->next;
+	regfree (&c_entry->preg);
+	if (c_entry->action.what == AC_EXECUTE)
 	{
 	    for (count = 0; count < MAX_ARGS; ++count)
 	    {
-		if (entry->u.execute.argv[count] == NULL) break;
-		free (entry->u.execute.argv[count]);
+		if (c_entry->u.execute.argv[count] == NULL) break;
+		free (c_entry->u.execute.argv[count]);
 	    }
 	}
-	free (entry);
+	else if ( (c_entry->action.what == AC_MFUNCTION) ||
+		  (c_entry->action.what == AC_CFUNCTION) )
+	{
+	    for (count = 0; count < MAX_ARGS; ++count)
+	    {
+		if (c_entry->u.function.argv[count] == NULL) break;
+		free (c_entry->u.function.argv[count]);
+	    }
+	}
+	free (c_entry);
     }
     first_config = NULL;
     last_config = NULL;
+    for (so_entry = first_so; so_entry != NULL; so_entry = next)
+    {
+	next = so_entry->next;
+	free ( (char *) so_entry->name );
+	if (so_entry->handle != RTLD_NEXT) dlclose (so_entry->handle);
+	free (so_entry);
+    }
+    first_so = NULL;
 }   /*  End Function free_config  */
 
 static void do_debug (int fd)
@@ -1437,7 +1605,8 @@ static uid_t get_uid (const char *string)
 {
     struct passwd *pw_ent;
 
-    if ( isdigit (string[0]) ) return atoi (string);
+    if ( isdigit (string[0]) || ( (string[0] == '-') && isdigit (string[1]) ) )
+	return atoi (string);
     if ( ( pw_ent = getpwnam (string) ) == NULL )
     {
 	SYSLOG (LOG_ERR, "unknown user: \"%s\"\n", string);
@@ -1455,7 +1624,8 @@ static gid_t get_gid (const char *string)
 {
     struct group *grp_ent;
 
-    if ( isdigit (string[0]) ) return atoi (string);
+    if ( isdigit (string[0]) || ( (string[0] == '-') && isdigit (string[1]) ) )
+	return atoi (string);
     if ( ( grp_ent = getgrnam (string) ) == NULL )
     {
 	SYSLOG (LOG_ERR, "unknown group: \"%s\"\n", string);
@@ -1573,7 +1743,7 @@ static void do_scan_and_service (DIR *dp, CONST char *dirname)
 	sprintf (path, "%s/%s", dirname, de->d_name);
 	if (lstat (path, &statbuf) != 0)
 	{
-	    SYSLOG (LOG_ERR, "Error stating file \"%s\"\t%s\n",
+	    SYSLOG (LOG_ERR, "Error stat(2)ing file \"%s\"\t%s\n",
 		    path, ERRSTRING);
 	    exit (RV_SYS_ERROR);
 	}
@@ -1645,3 +1815,31 @@ static flag make_symlink (const char *oldpath, const char *newpath)
     }
     return (TRUE);
 }   /*  End Function make_symlink  */
+
+static struct shared_object *get_shared_object (CONST char *name)
+/*  [SUMMARY] Load a shared object if possible.
+    <name> The shared object.
+    [RETURNS] A pointer to the object on success, else NULL.
+*/
+{
+    struct shared_object *so;
+
+    for (so = first_so; so != NULL; so = so->next)
+	if (strcmp (so->name, name) == 0) return (so);
+    if ( ( so = malloc (sizeof *so) ) == NULL ) return (NULL);
+    if ( ( so->name = strdup (name) ) == NULL )
+    {
+	free (so);
+	return (NULL);
+    }
+    if (strcmp (name, "GLOBAL") == 0) so->handle = RTLD_NEXT;
+    else if ( ( so->handle = dlopen (name, RTLD_NOW) ) == NULL )
+    {
+	free ( (char *) so->name );
+	free (so);
+	return (NULL);
+    }
+    so->next = first_so;
+    first_so = so;
+    return (so);
+}   /*  End Function get_shared_object  */
